require './exceptions'
# [["ld", ["a", 255]], ["nop", []], ["inc", ["b"]]]

MIN_ARGS = {
  nop: 0,
  ld: 2,
  ldd: 2,
  push: 1,
  add: 1,
  adc: 1,
  sub: 1,
  sbc: 1,
  and: 1,
  or: 1,
  xor: 1,
  cp: 1,
  inc: 1,
  dec: 1,
  swap: 1,
  daa: 0,
  cpl: 0,
  ccf: 0,
  scf: 0,
  halt: 0,
  stop: 0,
  di: 0,
  ei: 0,
  rlca: 0,
  rla: 0,
  rrca: 0,
  rra: 0,
  rlc: 1,
  rl: 1,
  rrc: 1,
  rr: 1,
  sla: 1,
  sra: 1,
  srl: 1,
  bit: 2,
  set: 2,
  res: 2,
  jp: 1,
  jr: 1,
  call: 1,
  rst: 1,
  ret: 0,
  reti: 0,
}

def int?(x)
  x.is_a?(Integer)
end

def invert_number(n)
  return n unless n > 0xFF

  up = (n & 0xFF00) >> 8
  low = n & 0xFF
  (low << 8) | up
end

def translate_s8(n)
  raise TranslatorException.new("Not a number: #{n}") unless int?(n)
  raise TranslatorException.new("Number too large: #{n}") if n < -128

  (0xFF + n + 1)
end

def translate_u16(n)
  raise TranslatorException.new("Not a number: #{n}") unless int?(n)
  raise TranslatorException.new("Number too large: #{n}") if n > 0xFFFF

  n = invert_number(n)
  up = (n & 0xFF00) >> 8
  low = n & 0xFF

  (up.chr + low.chr).b
end

def translate_number(n, max = -1)
  raise TranslatorException.new("Not a number: #{n}") unless int?(n)
  if n > max && max != -1
    raise TranslatorException.new("Number too large: #{n} (> #{max})")
  end

  if n < 0
    translate_s8(n).chr
  elsif n <= 0xFF
    n.chr
  elsif n <= 0xFFFF
    translate_u16(n)
  end
end

def translate_u8(n)
  translate_number(n, 0xFF)
end

def translate_absolute_deref(n)
  if n[0] != '(' || n[-1] != ')'
    raise TranslatorException.new("Syntax error: #{n}")
  end

  n = n[1..-2]

  # re-lex
  v = convert_numbers(n)
  raise TranslatorException.new("Syntax error (Not a number): #{n}") unless int?(v)

  translate_u16(v)
end

def ld_a(arg)
  case arg
  when 'a'
    "\x7f"
  when ('b'..'e')
    (0x78 + (arg.ord - 'b'.ord)).chr
  when 'h'
    "\x7c"
  when 'l'
    "\x7d"
  when '(bc)'
    "\x0a"
  when '(de)'
    "\x1a"
  when '(hl)'
    "\x7e"
  when '(hli)'
    "\x2a"
  when '(c)'
    "\xf2"
  else
    if int?(arg)
      "\x3e" + translate_u8(arg)
    elsif arg[0] == '(' && arg[-1] == ')'
      n = arg[1..-1].to_i
      if (0xFF00..0xFFFF).include?(n)
        "\xf0" + translate_u8(n & 0xFF)
      else
        "\xfa" + translate_absolute_deref(arg)
      end
    else
      raise TranslatorException.new("Unknown argument #{arg} to ld a")
    end
  end
end

# e.g. ld b,a or add a,b
def simple_r(base, arg)
  case arg
  when 'a'
    (base + 7).chr
  when ('b'..'e')
    (base + (arg.ord - 'b'.ord)).chr
  when 'h'
    (base + 4).chr
  when 'l'
    (base + 5).chr
  when '(hl)'
    (base + 6).chr
  else
    raise TranslatorException.new("Unknown argument #{arg}") unless ret
  end
end

def simple_r_wrap(base, arg, u8write)
  return u8write.b + translate_u8(arg) if int?(arg)
  simple_r(base, arg)
end

def ld_b(arg)
  simple_r_wrap(0x40, arg, "\x06")
end

def ld_c(arg)
  simple_r_wrap(0x48, arg, "\x0e")
end

def ld_d(arg)
  simple_r_wrap(0x50, arg, "\x16")
end

def ld_e(arg)
  simple_r_wrap(0x58, arg, "\x1e")
end

def ld_h(arg)
  simple_r_wrap(0x60, arg, "\x26")
end

def ld_l(arg)
  simple_r_wrap(0x68, arg, "\x2e")
end

def push_pop(base, arg)
  case arg
  when 'af'
    return (base + 0x30).chr
  when 'bc'
    return base.chr
  when 'de'
    return (base + 0x10).chr
  when 'hl'
    return (base + 0x20).chr
  else
    raise TranslatorException.new("Invalid argument (not one of af, bc, de, hl): #{arg}")
  end
end

def bit_set(args, base)
  n = args[0].to_i
  if (n == 0 && args[0] != '0' && args[0] != 0) || !(0..7).include?(n)
    raise TranslatorException.new("Invalid bit #{args[0]}")
  end

  return "\xcb".b + simple_r(base + n*8, args[1])
end

def translate_condition(base, arg)
  case arg
  when 'nz'
    return base.chr
  when 'z'
    return (base + 8).chr
  when 'nc'
    return (base + 16).chr
  when 'c'
    return (base + 24).chr
  else
    raise TranslatorException.new("Invalid condition (not one of nz, z, nc, c): #{arg}")
  end
end

def distance_for_label(label, inst_pos)
  target = $labels[label]
  if not target
    raise TranslatorException.new("No such label .#{label}")
  end
  diff = target - inst_pos

  if diff < -128 || diff > 127
    raise TranslatorException.new("Label .#{label} is more than -128/127 bytes away")
  end

  return (((0xFF + diff + (diff > 0 ? 1 : -1))) & 0xFF).chr
end

def translate_lexed(val)
  inst, args = val

  if inst.is_a?(Symbol)
    if $labels[inst]
      raise TranslatorException.new("Duplicate label: #{inst}")
    end

    $labels[inst] = $out.length
    return nil
  end

  raise TranslatorException.new('Too many arguments') if args.length > 2
  min_args = MIN_ARGS[inst.to_sym]
  raise TranslatorException.new("Insufficient arguments for #{inst} (required: #{min_args}, got: #{args.length})") if min_args > args.length

  # BIG translation table here. Ugh.
  case inst
  when 'nop'
    return "\x00"
  when 'ld'
    if args[0][0] == '('
      if args[0] == '(bc)'
        raise TranslatorException.new("Invalid argument to ld (bc) (not a): #{args[1]}") unless args[1] == 'a'
        return "\x02"
      elsif args[0] == '(de)'
        raise TranslatorException.new("Invalid argument to ld (de) (not a): #{args[1]}") unless args[1] == 'a'
        return "\x12"
      elsif args[0] == '(hli)'
        raise TranslatorException.new("Invalid argument to ld (hli) (not a): #{args[1]}") unless args[1] == 'a'
        return "\x22"
      elsif args[0] == '(hl)'
        return simple_r_wrap(0x70, args[1], "\x36")
      elsif args[0] == '(c)'
        raise TranslatorException.new("Invalid argument to ld (c) (not a): #{args[1]}") unless args[1] == 'a'
        return "\xe2"
      elsif args[1] == 'sp'
        return "\x08".b + translate_absolute_deref(args[0])
      elsif args[1] == 'a'
        n = args[0][1..-1].to_i
        if (0xFF00..0xFFFF).include?(n)
          return "\xe0".b + translate_u8(n & 0xFF)
        else
          return "\xea".b + translate_absolute_deref(args[0])
        end
      else
        raise TranslatorException.new("Invalid arguments to dereferencing ld: #{args.join(',')}")
      end
    end

    case args[0]
    when 'a'
      return ld_a(args[1])
    when 'b'
      return ld_b(args[1])
    when 'c'
      return ld_c(args[1])
    when 'd'
      return ld_d(args[1])
    when 'e'
      return ld_e(args[1])
    when 'h'
      return ld_h(args[1])
    when 'l'
      return ld_l(args[1])
    when 'bc'
      return "\x01" + translate_u16(args[1])
    when 'hl'
      if int?(args[1])
        return "\x21" + translate_u16(args[1])
      elsif args[1] == 'sp'
        return "\xf8"
      else
        raise TranslatorException.new("Invalid argument to ld hl (not a number and not sp): #{args[1]}")
      end
    when 'sp'
      if int?(args[1])
        return "\x31" + translate_u16(args[1])
      elsif args[1] == 'hl'
        return "\xf9"
      else
        raise TranslatorException.new("Invalid argument to ld sp (not a number and not sp): #{args[1]}")
      end
    else
      raise TranslatorException.new("Invalid target for ld: #{args[0]}")
    end
  when 'ldd'
    if args[0] == '(hl)' && args[1] == 'a'
      return "\x32"
    elsif args[0] == 'a' && args[1] == '(hl)'
      return "\x3a"
    else
      raise TranslatorException.new("Invalid argument(s) to ldd (not any of (hl) and a): #{args.join(',')}")
    end
  when 'push'
    return push_pop(0xc5, args[0])
  when 'pop'
    return push_pop(0xc1, args[0])
  when 'add'
    args.unshift('a') if args.length < 2

    case args[0]
    when 'a'
      return simple_r_wrap(0x80, args[1], "\xc6")
    when 'hl'
      case args[1]
      when 'bc'
        return "\x09"
      when 'de'
        return "\x19"
      when 'hl'
        return "\x29"
      when 'sp'
        return "\x39"
      end
    when 'sp'
      raise TranslatorException.new("Not a number: #{args[1]}") unless int?(args[1])
      return "\xe8" + translate_number(args[1], 127)
    else
      raise TranslatorException.new("Invalid target for add: #{args[0]}")
    end
  when 'adc'
    args.unshift('a') if args.length < 2

    raise TranslatorException.new("Invalid target for adc (not a): #{args[0]}") unless args[0] == 'a'

    return simple_r_wrap(0x88, args[1], "\xce")
  when 'sub'
    args.unshift('a') if args.length < 2

    raise TranslatorException.new("Invalid target for sub (not a): #{args[0]}") unless args[0] == 'a'

    return simple_r_wrap(0x90, args[1], "\xd6")
  when 'sbc'
    args.unshift('a') if args.length < 2

    raise TranslatorException.new("Invalid target for sbc (not a): #{args[0]}") unless args[0] == 'a'

    return simple_r_wrap(0x98, args[1], "\xde")
  when 'and'
    return simple_r_wrap(0xa0, args[0], "\xe6")
  when 'or'
    return simple_r_wrap(0xb0, args[0], "\xf6")
  when 'xor'
    return simple_r_wrap(0xa8, args[0], "\xee")
  when 'cp'
    return simple_r_wrap(0xb8, args[0], "\xfe")
  when 'inc'
    case args[0]
    when 'a'
      return "\x3c"
    when 'bc'
      return "\x03"
    when 'b'
      return "\x04"
    when 'c'
      return "\x0c"
    when 'de'
      return "\x13"
    when 'd'
      return "\x14"
    when 'e'
      return "\x1c"
    when 'hl'
      return "\x23"
    when 'h'
      return "\x24"
    when 'l'
      return "\x2c"
    when 'sp'
      return "\x33"
    when '(hl)'
      return "\x34"
    else
      raise TranslatorException.new("Invalid target for inc (not a, bc, b, c, de, d, e, hl, h, l, sp, (hl)): #{args[0]}")
    end
  when 'dec'
    case args[0]
    when 'a'
      return "\x3d"
    when 'bc'
      return "\x0b"
    when 'b'
      return "\x05"
    when 'c'
      return "\x0d"
    when 'de'
      return "\x1b"
    when 'd'
      return "\x15"
    when 'e'
      return "\x1d"
    when 'hl'
      return "\x2b"
    when 'h'
      return "\x25"
    when 'l'
      return "\x2d"
    when 'sp'
      return "\x3b"
    when '(hl)'
      return "\x35"
    else
      raise TranslatorException.new("Invalid target for dec (not a, bc, b, c, de, d, e, hl, h, l, sp, (hl)): #{args[0]}")
    end
  when 'swap'
    return "\xcb" + simple_r(0x30, args[0])
  when 'daa'
    return "\x27"
  when 'cpl'
    return "\x2f"
  when 'ccf'
    return "\x3f"
  when 'scf'
    return "\x37"
  when 'halt'
    return "\x76"
  when 'stop'
    return "\x10\x00"
  when 'di'
    return "\xf3"
  when 'ei'
    return "\xfb"
  when 'rlca'
    return "\x07"
  when 'rla'
    return "\x17"
  when 'rrca'
    return "\x0f"
  when 'rra'
    return "\x1f"
  when 'rlc'
    return "\xcb" + simple_r(0x00, args[0])
  when 'rl'
    return "\xcb" + simple_r(0x10, args[0])
  when 'rrc'
    return "\xcb" + simple_r(0x08, args[0])
  when 'rr'
    return "\xcb" + simple_r(0x18, args[0])
  when 'sla'
    return "\xcb" + simple_r(0x20, args[0])
  when 'sra'
    return "\xcb" + simple_r(0x28, args[0])
  when 'srl'
    return "\xcb" + simple_r(0x38, args[0])
  when 'bit'
    return bit_set(args, 0x40)
  when 'set'
    return bit_set(args, 0xc0)
  when 'res'
    return bit_set(args, 0x80)
  when 'jp'
    if args.length == 1
      if args[0] == '(hl)'
        return "\xe9"
      else
        return "\xc3" + translate_u16(args[0])
      end
    end

    opcode = translate_condition(0x20, args[0])

    return opcode + translate_u16(args[1])
  when 'jr'
    if args.length == 1
      # Wait for pass 2 in case of labels
      if args[0].is_a?(Symbol)
        if $label_uses[args[0]]
          $label_uses[args[0]] << ["\x18", $out.length]
        else
          $label_uses[args[0]] = [["\x18", $out.length]]
        end
        # Pad to keep offsets stable
        return "\x18\x00"
      end

      return "\x18" + translate_number(args[0], 127)
    end

    opcode = translate_condition(0x20, args[0])

    if args[1].is_a?(Symbol)
      if $label_uses[args[1]]
        $label_uses[args[1]] << [opcode, $out.length]
      else
        $label_uses[args[1]] = [[opcode, $out.length]]
      end

      # Pad to keep offsets stable
      return opcode + "\x00"
    end

    return opcode + translate_number(args[1], 127)
  when 'call'
    if args.length == 1
      return "\xcd".b + translate_u16(args[0])
    end

    case args[0]
    when 'nz'
      return "\xc4".b + translate_u16(args[1])
    when 'z'
      return "\xcc".b + translate_u16(args[1])
    when 'nc'
      return "\xd4".b + translate_u16(args[1])
    when 'c'
      return "\xdc".b + translate_u16(args[1])
    else
      raise TranslatorException.new("Invalid condition (not one of nz, z, nc, c): #{args[0]}")
    end
  when 'rst'
    n = args[0].to_i
    if n == 0 && args[0] != '0' && args[0] != 0
      raise TranslatorException.new("Unknown rst value (not a number): #{args[0]}")
    end

    case n
    when 0x00
      return "\xc7"
    when 0x08
      return "\xcf"
    when 0x10
      return "\xd7"
    when 0x18
      return "\xdf"
    when 0x20
      return "\xe7"
    when 0x28
      return "\xef"
    when 0x30
      return "\xf7"
    when 0x38
      return "\xff"
    else
      raise TranslatorException.new("Unknown rst value (not one of 0x00, 0x08, 0x10, 0x18, 0x20, 0x28, 0x30, 0x38): #{n}")
    end
  when 'ret'
    if args.length == 0
      return "\xc9"
    end

    return translate_condition(0xc0, args[0])
  when 'reti'
    return "\xd9"
  end
end

def translate(vals)
  errors = []
  $labels = {}
  $out = ''.b
  $label_uses = {}

  vals.each do |val|
    begin
      res = translate_lexed(val)
      $out += res.b if res
    rescue TranslatorException => e
      errors << "Error at #{val.flatten.join(' ')}: #{e.message}"
    end
  end

  if !errors.empty?
    puts errors.join("\n") + "\nErrors found, aborting."
    exit(1)
  end

  $labels.each do |label, label_pos|
    $label_uses[label].each do |op|
      opcode, inst_pos = op

      begin
        $out[inst_pos + 1] = distance_for_label(label, inst_pos)
      rescue TranslatorException => e
        errors << "Error at #{val.flatten.join(' ')}: #{e.message}"
      end
    end
  end

  if !errors.empty?
    puts errors.join("\n") + "\nErrors found, aborting."
    exit(1)
  end

  $out
end

